import torch
import torch.nn as nn
import logging
import numpy as np
import torch.optim as optim

from ResCSNet.resCSnet_model import ResCsNet,ResCsNet_init_weights
from ResCSNet.pytorch_dwt_ssim import FrobeniusLoss, DWT_SSIM
from ResCSNet.LoadCOCO import coco_dataset

exp_name = 'ResCsNet-colab'

                  
def get_evaluation(model, loader_val, loader_subtrain, device, fn_mse, fn_cwssim, mse_weight=0.3, cwssim_weight=0.7):
    '''
    Evaluate the subtrain set and validation set, returns
    subtrain_psnr, val_psnr, subtrain_mix_psnr, val_mix_psnr
    '''
    model.eval()  # ensure the model is in evaluation mode
    subtrain_avg_psnr = 0.0
    subtrain_avg_mse = 0.0
    val_avg_psnr = 0.0
    val_avg_mse = 0.0
    
    fn_mse = fn_mse.to(device=device)
    fn_cwssim = fn_cwssim.to(device=device)
    subtrain_avg_cwssim_loss = 0.0
    val_avg_cwssim_loss = 0.0
    
    with torch.no_grad():
        for _, val_im in enumerate(loader_val):
            val_original = val_im.to(device)
            val_recovered = model(val_original)
            val_mse = fn_mse(val_recovered, val_original)
            val_avg_mse += val_mse
            # PSNR
            val_psnr = -10 * torch.log10(val_mse)
            val_avg_psnr += val_psnr
            # CW-SSIM
            val_original = val_im.to(device)
            val_recovered = model(val_original)
            val_cwssim_loss = 1 - fn_cwssim(val_recovered, val_original)
            val_avg_cwssim_loss += val_cwssim_loss
            
        val_avg_mse /= len(loader_val)
        val_avg_psnr /= len(loader_val)
        val_avg_cwssim_loss /= len(loader_val)

        for _, subtrain_im in enumerate(loader_subtrain):
            subtrain_original = subtrain_im.to(device)
            subtrain_recovered = model(subtrain_original)
            subtrain_mse = fn_mse(subtrain_recovered, subtrain_original)
            subtrain_avg_mse += subtrain_mse
            # PSNR
            subtrain_psnr = -10 * torch.log10(subtrain_mse)
            subtrain_avg_psnr += subtrain_psnr
            # CW-SSIM
            subtrain_original = subtrain_im.to(device)
            subtrain_recovered = model(subtrain_original)
            subtrain_cwssim_loss = 1 - fn_cwssim(subtrain_recovered, subtrain_original)
            subtrain_avg_cwssim_loss += subtrain_cwssim_loss

        subtrain_avg_mse /= len(loader_subtrain)
        subtrain_avg_psnr /= len(loader_subtrain)
        subtrain_avg_cwssim_loss /= len(loader_subtrain)
    
    subtrain_psnr = subtrain_avg_psnr
    val_psnr = val_avg_psnr
    subtrain_mix_psnr = -10 * torch.log10(mse_weight*subtrain_avg_mse + cwssim_weight*subtrain_avg_cwssim_loss)
    val_mix_psnr = -10 * torch.log10(mse_weight*val_avg_mse + cwssim_weight*val_avg_cwssim_loss)
    return subtrain_psnr, val_psnr, subtrain_mix_psnr, val_mix_psnr

def train_Epoches(model, optimizer, lr_scheduler, epochs, device, loader_train):
    """
    Inputs:
    - model: A PyTorch Module giving the model to train.
    - optimizer: An Optimizer object we will use to train the model
    - lr_scheduler: learning rate scheduler
    - fn_mse, fn_cwssim: loss functions for L2 loss and wavelet loss
    - mse_weight, cwssim_weight: weights indicating how important they contribute to the total loss
    - epochs: A Python integer giving the number of epochs to train for
    - logdir: string. Used to specific the logdir of tensorboard
    - print_every: after print_every epochs this function will report to logging
    - device: torch.device('cuda') or torch.device('cpu')

    Returns:
    - tfx_steps: the end of the tfx_steps
    """

    fn_mse=FrobeniusLoss()
    fn_cwssim=DWT_SSIM(J=3, wave='haar')
    fn_mse = fn_mse.to(device=device)
    fn_cwssim = fn_cwssim.to(device=device)
    mse_weight=0.3
    cwssim_weight=0.7
    print_every=20
    # PSNR scheduler
#         lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='max', factor=0.2, patience=10, threshold=5e-3 ,verbose=True)
    # CW-SSIM scheduler
#     lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='max', patience=10, verbose=True)
    
    for e in range(epochs):        
        logging.info('-----------------------------')
        for t, original_im in enumerate(loader_train):
            original_im = original_im.to(device=device)
                        
            recovered_im = model(original_im)    
                    
            mse_loss = fn_mse(recovered_im, original_im)
            cwssim_loss = 1 - fn_cwssim(recovered_im, original_im)
            # construct the total loss
            loss = mse_weight*mse_loss + cwssim_weight*cwssim_loss

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            if t % int(print_every) == 0:
                print('Epoch %d/%d, Iteration %d/%d, loss = %.4f' % (e,epochs, t, len(loader_train) , loss.item()))
    #end for


    ## check the performance of the model
    print("Checking on subtrain and validation set...")
    subtrain_psnr, val_psnr, subtrain_mix_psnr, val_mix_psnr = get_evaluation(
        model,
        loader_val,
        loader_subtrain,
        device,
        fn_mse=nn.MSELoss(),
        fn_cwssim=fn_cwssim,
        mse_weight=mse_weight,
        cwssim_weight=cwssim_weight)
    print("Average PSNR for subtrain set is", subtrain_psnr, "dB")
    print("Average PSNR for validation set is", val_psnr, "dB")

    print("Average mixed gain for subtrain set is", subtrain_mix_psnr, "dB")
    print("Average mixed gain for validation set is", val_mix_psnr, "dB")

#         loss_subtrain = mse_weight*subtrain_avg_mse + cwssim_weight*subtrain_avg_cwssim_loss
#         loss_val = mse_weight*val_avg_mse + cwssim_weight*val_avg_cwssim_loss
    if lr_scheduler is not None:
        lr_scheduler.step(val_mix_psnr)  # check loss and determine if the lr should be decreased

    #end for           
   
    print("Saving the trained model")
    # Save the state_dict of the model
    torch.save(model.state_dict(), exp_name+"_0.10.pth")



    
def train_ResCSNet(loader_train,SAMPLING_RATE, input_width_height,epochs = 3):
    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    print('using device:', device)

    N = input_width_height*input_width_height
    M = int(N * SAMPLING_RATE)
    print("N,M ==", N, M)
    model = ResCsNet(N, M)
    model.apply(ResCsNet_init_weights)  # not needed if continue trainning or testing
    model.train()  # ensure the model is in training mode
    model.to(device)
    
    optimizer = optim.Adam(model.parameters())   
    lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='max', factor=0.2, patience=10, threshold=5e-3 ,verbose=True)
    
    train_Epoches(model, optimizer, lr_scheduler, epochs, device, loader_train)
                       
if __name__ == '__main__':
    
    input_width_height = 120
    print('input_width_height:', input_width_height)
    
    loader_train, loader_subtrain, loader_val, loader_test = coco_dataset(input_width_height)

    SAMPLING_RATE = 0.10
    epochs = 300
    modelname = '/home/chengzi/Desktop/workspace20170624/CSMRI/ResCSNet/ResCsNet-colab-5_2_1-r0.10_checkpoint.pth'
    
    train_ResCSNet(loader_train,SAMPLING_RATE,input_width_height, epochs)
